import os
import logging

def make_logger(name):
    logger = logging.getLogger(name)
    handler = logging.NullHandler()
    formatter = logging.Formatter(
            '%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    if os.environ.get('AHK_DEBUG'):
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.ERROR)
    return logger



from collections import deque
from ahk.mouse import MouseMixin
from ahk.window import Window, WindowMixin
from ahk.script import ScriptEngine
from ahk.screen import ScreenMixin

class AHK(WindowMixin, MouseMixin, ScreenMixin):
    pass


class Hotkey:
    def __init__(self, engine: ScriptEngine, hotkey, script):
        self.hotkey = hotkey
        self.script = script
        self.engine = engine

    @property
    def running(self):
        return hasattr(self, '_proc')

    def _start(self, script):
        try:
            proc = self.engine.run_script(script, blocking=False)
            yield proc
        finally:
            self._stop()

    def start(self):
        if self.running:
            raise RuntimeError('Hotkey is already running')
        script = self.engine.render_template('hotkey.ahk', blocking=False, script=self.script, hotkey=self.hotkey)
        self._gen = self._start(script)
        proc = next(self._gen)
        self._proc = proc

    def _stop(self):
        if not self.running:
            return
        self._proc.terminate()
        del self._proc

    def stop(self):
        if not self.running:
            return
        try:
            next(self._gen)
        except StopIteration:
            pass
        finally:
            del self._gen


class ActionChain(AHK):
    def __init__(self, *args, **kwargs):
        self._actions = deque()
        super().__init__(*args, **kwargs)

    def run_script(self, *args, **kwargs):
        """
        override run_script to defer to queue
        """
        kwargs['decode'] = False  #
        self._actions.appendleft((args, kwargs))

    def perform(self):
        results = []
        while self._actions:
            args, kwargs = self._actions.pop()
            results.append(super().run_script(*args, **kwargs))
        return results

    def sleep(self, n):
        """
        :param n: how long (in seconds) to sleep
        :return:
        """
        n = n * 1000  # convert to miliseconds
        script = self.render_template('base.ahk', body=f'Sleep {n}', directives={'#Persistent',})
        self.run_script(script)



from collections import namedtuple
from ahk.script import ScriptEngine
from ahk.utils import make_logger
import ast

logger = make_logger(__name__)


_BUTTONS = {
    1: 'L',
    2: 'R',
    3: 'M',
    'left': 'L',
    'right': 'R',
    'middle': 'M',
    'wheelup': 'WU',
    'wheeldown': 'WD',
    'wheelleft': 'WL',
    'wheelright': 'WR',
}

def resolve_button(button):
    if isinstance(button, str):
        button = button.lower()

    if button in _BUTTONS:
        button = _BUTTONS.get(button)
    elif isinstance(button, int) and button > 3:
        #  for addtional mouse buttons
        button = f'X{button-3}'
    return button


class MouseMixin(ScriptEngine):
    def __init__(self, mouse_speed=2, mode=None, **kwargs):
        if mode is None:
            mode = 'Screen'
        self.mode = mode
        self._mouse_speed = mouse_speed
        super().__init__(**kwargs)

    @property
    def mouse_speed(self):
        if callable(self._mouse_speed):
            return self._mouse_speed()
        else:
            return self._mouse_speed

    @mouse_speed.setter
    def mouse_speed(self, value):
        self._mouse_speed = value

    def _mouse_position(self, mode=None):
        if mode is None:
            mode = self.mode
        return self.render_template('mouse/mouse_position.ahk', mode=mode)

    @property
    def mouse_position(self):
        script = self._mouse_position()
        response = self.run_script(script)
        return ast.literal_eval(response)

    @mouse_position.setter
    def mouse_position(self, position):
        x, y = position
        self.mouse_move(x=x, y=y, speed=0, relative=False)

    def _mouse_move(self, x=None, y=None, speed=None, relative=False, mode=None, blocking=True):
        if x is None and y is None:
            raise ValueError('Position argument(s) missing. Must provide x and/or y coordinates')
        if speed is None:
            speed = self.mouse_speed
        if mode is None:
            mode = self.mode
        if relative and (x is None or y is None):
            x = x or 0
            y = y or 0
        elif not relative and (x is None or y is None):
            posx, posy = self.mouse_position
            x = x or posx
            y = y or posy

        return self.render_template('mouse/mouse_move.ahk', x=x, y=y, speed=speed, relative=relative, mode=mode, blocking=blocking)

    def mouse_move(self, *args, **kwargs):
        blocking = kwargs.get('blocking', True)
        script = self._mouse_move(*args, **kwargs)
        self.run_script(script, blocking=blocking)

    def _click(self, *args, mode=None, blocking=True):
        if mode is None:
            mode = self.mode
        return self.render_template('mouse/click.ahk', args=args, mode=mode, blocking=blocking)

    def click(self, x=None, y=None, *, button=None, n=None, direction=None, relative=None, blocking=True, mode=None):
        if x or y:
            if y is None and not isinstance(x, int) and len(x) == 2:
                #  alow position to be specified by a two-sequence
                x, y = x
            assert x is not None and y is not None, 'If provided, position must be specified by x AND y'

        button = resolve_button(button)

        if relative:
            relative = 'Rel'
        args = [arg for arg in (x, y, button, n, direction, relative) if arg is not None]
        script = self._click(*args, blocking=blocking, mode=mode)
        self.run_script(script, blocking=blocking)

    def double_click(self, *args, **kwargs):
        n = kwargs.get('n', 1)
        kwargs['n'] = n * 2
        self.click(*args, **kwargs)

    def right_click(self, *args, **kwargs):
        kwargs['button'] = 2
        self.click(*args, **kwargs)

    def mouse_wheel(self, direction, *args, **kwargs):
        assert direction in ('up', 'down')
        kwargs['button'] = f'Wheel{direction}'
        self.click(*args, **kwargs)

    def wheel_up(self, *args, **kwargs):
        self.mouse_wheel('up', *args, **kwargs)

    def wheel_down(self, *args, **kwargs):
        self.mouse_wheel('down', *args, **kwargs)

    def mouse_drag(self, x, y=None, *, from_position=None, speed=None, button=1, relative=None, blocking=True, mode=None):
        if from_position is None:
            x1, y1 = self.mouse_position
        else:
            x1, y1 = from_position

        if y is None:
            x2, y2 = x
        else:
            x2 = x
            y2 = y

        if relative:
            x1, y1 = (0, 0)

        button = resolve_button(button)

        if speed is None:
            speed = self.mouse_speed

        if mode is None:
            mode = self.mode

        script = self.render_template('mouse/mouse_drag.ahk',
                                      button=button,
                                      x1=x1,
                                      y1=y1,
                                      x2=x2,
                                      y2=y2,
                                      speed=speed,
                                      relative=relative,
                                      blocking=blocking,
                                      mode=mode)

        self.run_script(script, blocking=blocking)



import os
import subprocess
from shutil import which
from ahk.utils import make_logger
from ahk.directives import Persistent
from jinja2 import Environment, FileSystemLoader

logger = make_logger(__name__)


class ExecutableNotFoundError(EnvironmentError):
    pass


class ScriptEngine(object):
    def __init__(self, executable_path: str='', **kwargs):
        """
        :param executable_path: the path to the AHK executable.
        Defaults to environ['AHK_PATH'] if not explicitly provided
        If environment variable not present, tries to look for 'AutoHotkey.exe' or 'AutoHotkeyA32.exe' with shutil.which
        :param keep_scripts:
        :raises ExecutableNotFound: if AHK executable is not provided and cannot be found in environment variables or PATH
        """
        if not executable_path:
            executable_path = os.environ.get('AHK_PATH') or which('AutoHotkey.exe') or which('AutoHotkeyA32.exe')
        if not executable_path:
            raise ExecutableNotFoundError('Could not find AutoHotkey.exe on PATH. '
                                          'Provide the absolute path with the `executable_path` keyword argument '
                                          'or in the AHK_PATH environment variable.')
        self.executable_path = executable_path
        templates_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'templates')
        self.env = Environment(
            loader=FileSystemLoader(templates_path),
            autoescape=False,
            trim_blocks=True
        )

    def render_template(self, template_name, directives=None, blocking=True, **kwargs):
        if directives is None:
            directives = set()
        else:
            directives = set(directives)
        if blocking:
            directives.add(Persistent)
        elif Persistent in directives:
            directives.remove(Persistent)

        kwargs['directives'] = directives
        template = self.env.get_template(template_name)
        return template.render(**kwargs)

    def _run_script(self, script_text, **kwargs):
        blocking = kwargs.pop('blocking', True)
        runargs = [self.executable_path, '/ErrorStdOut', '*']
        decode = kwargs.pop('decode', False)
        script_bytes = bytes(script_text, 'utf-8')
        if blocking:
            result = subprocess.run(runargs, input=script_bytes, stderr=subprocess.PIPE, stdout=subprocess.PIPE, **kwargs)
            if decode:
                logger.debug('Stdout: %s', repr(result.stdout))
                logger.debug('Stderr: %s', repr(result.stderr))
                return result.stdout.decode()
            else:
                return result
        else:
            proc = subprocess.Popen(runargs, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs)
            try:
                proc.communicate(script_bytes, timeout=0)
            except subprocess.TimeoutExpired:
                pass  # for now, this seems needed to avoid blocking and use stdin
            return proc

    def run_script(self, script_text: str, decode=True, blocking=True, **runkwargs):
        logger.debug('Running script text: %s', script_text)
        try:
            result = self._run_script(script_text, decode=decode, blocking=blocking, **runkwargs)
        except Exception as e:
            logger.fatal('Error running temp script: %s', e)
            raise
        return result



from ahk.script import ScriptEngine
import ast
from .utils import make_logger

logger = make_logger(__name__)


class WindowNotFoundError(ValueError):
    pass

class Control:
    def __init__(self):
        raise NotImplementedError

    def click(self):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlClick.htm
        :return:
        """
        raise NotImplementedError

    def focus(self):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlFocus.htm
        :return:
        """
        raise NotImplementedError

    def get(self, key):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlGet.htm
        :param key:
        :return:
        """
        raise NotImplementedError

    def has_focus(self):
        raise NotImplementedError

    @property
    def position(self):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlGetPos.htm
        :return:
        """
        raise NotImplementedError

    @property
    def text(self):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlGetText.htm
        :return:
        """
        raise NotImplementedError

    @text.setter
    def text(self, new_text):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlSetText.htm
        :param new_text:
        :return:
        """
        raise NotImplementedError

    def move(self):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlMove.htm
        :return:
        """
        raise NotImplementedError

    def send(self, raw=False):
        """
        REF: https://www.autohotkey.com/docs/commands/ControlSend.htm
        :param raw:
        :return:
        """
        raise NotImplementedError




class Window(object):
    _subcommands = {
        'id': 'ID',
        'process_name': 'ProcessName',
        'pid': 'PID',
        'process_path': 'ProcessPath',
        'process': 'ProcessPath',
        'controls': 'ControlList',
        'controls_hwnd': 'ControlListHwnd',
        'transparency': 'Transparent',
        'trans_color': 'TransColor',
        'style': 'Style',   # This will probably get a property later
        'ex_style': 'ExStyle',  # This will probably get a property later
    }
    #  add reverse lookups
    _subcommands.update({value: value for value in _subcommands.values()})

    def __init__(self, engine: ScriptEngine, ahk_id: str, encoding=None):
        self.engine = engine  # should this be a weakref instead?
        self.id = ahk_id
        self.encoding = encoding

    @classmethod
    def from_mouse_position(cls, engine: ScriptEngine, **kwargs):
        script = engine.render_template('window/from_mouse.ahk')
        ahk_id = engine.run_script(script)
        return cls(engine=engine, ahk_id=ahk_id, **kwargs)

    @classmethod
    def from_pid(cls, engine: ScriptEngine, pid, **kwargs):
        script = engine.render_template('window/get.ahk',
                                        subcommand="ID",
                                        title=f'ahk_pid {pid}')
        ahk_id = engine.run_script(script)
        return cls(engine=engine, ahk_id=ahk_id, **kwargs)

    def __getattr__(self, attr):
        if attr.lower() in self._subcommands:
            return self.get(attr)
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{attr}'")

    def get(self, subcommand):
        sub = self._subcommands.get(subcommand)
        if not sub:
            raise ValueError(f'No such subcommand {subcommand}')
        script = self._render_template('window/get.ahk',
                                       subcommand=sub,
                                       title=f'ahk_id {self.id}')

        return self.engine.run_script(script)

    def __repr__(self):
        return f'<ahk.window.Window ahk_id={self.id}>'

    def win_set(self, subcommand, value):
        script = self._render_template('window/win_set.ahk', subcommand=subcommand, value=value)
        self.engine.run_script(script)

    def _get_pos(self):
        script = self._render_template('window/win_position.ahk')
        resp = self.engine.run_script(script)
        try:
            value = ast.literal_eval(resp)
            return value
        except SyntaxError:
            raise WindowNotFoundError('No window found')

    @property
    def rect(self):
        return self._get_pos()

    @rect.setter
    def rect(self, new_position):
        x, y, width, height = new_position
        self.move(x=x, y=y, width=width, height=height)

    @property
    def position(self):
        x, y, _, _ = self._get_pos()

        return x, y

    @position.setter
    def position(self, new_position):
        x, y = new_position
        self.move(x, y)

    @property
    def width(self):
        _, _, width, _ = self._get_pos()
        return width

    @width.setter
    def width(self, new_width):
        self.move(width=new_width)

    @property
    def height(self):
        _, _, _, height = self._get_pos()
        return height

    @height.setter
    def height(self, new_height):
        self.move(height=new_height)

    def disable(self):
        self.win_set('Disable', '')

    def enable(self):
        self.win_set('Enable', '')

    def redraw(self):
        self.win_set('Redraw', '')

    @property
    def title(self):
        script = self._render_template('window/win_get_title.ahk')
        result = self.engine.run_script(script, decode=False)
        if self.encoding:
            return result.stdout.decode(encoding=self.encoding)
        return result.stdout

    @property
    def text(self):
        script = self._render_template('window/win_get_text.ahk')
        result = self.engine.run_script(script, decode=False)
        if self.encoding:
            return result.stdout.decode(encoding=self.encoding)
        return result.stdout

    @property
    def always_on_top(self):
        script = self.render_template('window/win_is_always_on_top.ahk')
        resp = self.engine.run_script(script)
        return bool(ast.literal_eval(resp))

    @always_on_top.setter
    def always_on_top(self, value):
        if value in ('on', 'On', True, 1):
            self.win_set('AlwaysOnTop', 'On')
        elif value in ('off', 'Off', False, 0):
            self.win_set('AlwaysOnTop', 'Off')
        elif value in ('toggle', 'Toggle', -1):
            self.win_set('AlwaysOnTop', 'Toggle')
        else:
            raise ValueError(f'"{value}" not a valid option. Please use On/Off/Toggle/True/False/0/1/-1')

    def close(self, seconds_to_wait=''):
        script = self._render_template('window/win_close.ahk', seconds_to_wait=seconds_to_wait)
        self.engine.run_script(script)

    def to_bottom(self):
        """
        Send window to bottom (behind other windows)
        :return:
        """
        self.win_set('Bottom', '')

    def to_top(self):
        self.win_set('Top', '')

    def _render_template(self, *args, **kwargs):
        kwargs['win'] = self
        return self.engine.render_template(*args, **kwargs)

    def activate(self):
        script = self._render_template('window/win_activate.ahk')
        self.engine.run_script(script)

    def move(self, x='', y='', width=None, height=None):
        script = self._render_template('window/win_move.ahk', x=x, y=y, width=width, height=height)
        self.engine.run_script(script)

    def __eq__(self, other):
        if not isinstance(other, Window):
            return False
        return self.id == other.id

    def __hash__(self):
        return hash(repr(self))

class WindowMixin(ScriptEngine):
    def __init__(self, *args, **kwargs):
        self.window_encoding = kwargs.pop('window_encoding', None)
        super().__init__(*args, **kwargs)

    def win_get(self, title='', text='', exclude_title='', exclude_text='', encoding=None):
        encoding = encoding or self.window_encoding
        script = self.render_template('window/get.ahk',
                                      subcommand='ID',
                                      title=title,
                                      text=text,
                                      exclude_text=exclude_text,
                                      exclude_title=exclude_title)
        ahk_id = self.run_script(script)
        return Window(engine=self, ahk_id=ahk_id, encoding=encoding)

    def win_set(self, subcommand, *args, blocking=True):
        script = self.render_template('window/set.ahk',  subcommand=subcommand, *args, blocking=blocking)
        self.run_script(script, blocking=blocking)

    @property
    def active_window(self):
        return self.win_get(title='A')

    def _all_window_ids(self):
        script = self.render_template('window/id_list.ahk')
        result = self.run_script(script)
        return result.split('\n')[:-1]  # last one is always an empty string

    def windows(self):
        """
        Returns a list of windows
        :return:
        """
        windowze = []
        for ahk_id in self._all_window_ids():
            win = Window(engine=self, ahk_id=ahk_id, encoding=self.window_encoding)
            windowze.append(win)
        return windowze

    def find_windows(self, func=None, **kwargs):
        if func is None:
            exact = kwargs.pop('exact', False)
            def func(win):
                for attr, expected in kwargs.items():
                    if exact:
                        result = getattr(win, attr) == expected
                    else:
                        result = expected in getattr(win, attr)
                    if result is False:
                        return False
                return True
        for window in filter(func, self.windows()):
            yield window

    def find_window(self, func=None, **kwargs):
        return next(self.find_windows(func=func, **kwargs))

    def find_windows_by_title(self, title, exact=False):
        for window in self.find_windows(title=title, exact=exact):
            yield window

    def find_window_by_title(self, *args, **kwargs):
        return next(self.find_windows_by_title(*args, **kwargs))

    def find_windows_by_text(self, text, exact=False):
        for window in self.find_windows(text=text, exact=exact):
            yield window

    def find_window_by_text(self, *args, **kwargs):
        return next(self.find_windows_by_text(*args, **kwargs))



import ast
from ahk.script import ScriptEngine
from typing import Tuple, Union, Optional


class ScreenMixin(ScriptEngine):
    def image_search(self, image_path: str,
                     upper_bound: Tuple[int, int]=(0, 0), lower_bound: Tuple[int, int]=None,
                     coord_mode: str='Screen',
                     scale_height: int=None, scale_width: int=None) -> Union[Tuple[int, int], None]:
        """
        `AutoHotkey ImageSearch reference`_

        .. _AutoHotkey ImageSearch reference: https://autohotkey.com/docs/commands/ImageSearch.htm


        :param image_path: path to the image file e.g. C:\location\of\cats.png
        :param upper_bound: a two-tuple of X,Y coordinates for the upper-left corner of the search area e.g. (200, 400)
        defaults to (0,0)

        :param lower_bound: like ``upper_bound`` but for the lower-righthand corner of the search area e.g. (400, 800)
        defaults to screen width and height (lower right-hand corner; ``%A_ScreenWidth%``, ``%A_ScreenHeight%``).

        :param coord_mode: the Pixel CoordMode to use. Default is 'Screen'
        :param scale_height: Scale height in pixels. Equivalent of ``*hn`` option
        :param scale_width: Scale width in pixels. Equivalent of ``*wn`` option

        :return: coordinates of the upper-left pixel of where the image was found on the screen; ``None`` if the image
        was not found

        :rtype: Union[Tuple[int, int], None]

        Note: when only scale_height or only scale_width are provided, aspect ratio is maintained by default.
        """

        if scale_height and not scale_width:
            scale_width = -1
        elif scale_width and not scale_height:
            scale_height = -1

        x1, y1 = upper_bound
        if lower_bound:
            x2, y2 = lower_bound
        else:
            x2, y2 = ('%A_ScreenWidth%', '%A_ScreenHeight%')
        script = self.render_template('screen/image_search.ahk',
                                      x1=x1, x2=x2, y1=y1, y2=y2,
                                      coord_mode=coord_mode,
                                      scale_width=scale_width,
                                      scale_height=scale_height,
                                      image_path=image_path)
        resp = self.run_script(script)
        try:
            return ast.literal_eval(resp)
        except SyntaxError:
            return None

    def pixel_get_color(self, x: int, y: int, coord_mode: str='Screen',
                        alt: bool=False, slow: bool=False, rgb=True) -> Union[str, None]:
        """
        `AutoHotkey PixelGetColor reference`_

        .. _AutoHotkey PixelGetColor reference: https://autohotkey.com/docs/commands/PixelGetColor.htm

        :param x: x coordinate
        :param y: y coordinate
        :param coord_mode:
        :param alt:
        :param slow:
        :param rgb: returns
        :return: the color as an RGB hexidecimal string;
        :rtype: str
        """

        options = []
        if slow:
            options.append('Slow')
        elif alt:
            options.append('Alt')
        if rgb:
            options.append('RGB')
        script = self.render_template('screen/pixel_get_color.ahk',
                                      x=x, y=y,
                                      coord_mode=coord_mode,
                                      options=options)
        resp = self.run_script(script)
        return resp

    def pixel_search(self, color: Union[str, int], variation: int=0,
                     upper_bound: Tuple[int, int]=(0, 0), lower_bound: Tuple[int, int]=None,
                     coord_mode: str='Screen', fast: bool=True, rgb: bool=True) -> Union[Tuple[int, int], None]:
        """
        `AutoHotkey PixelSearch reference`_

        .. _AutoHotkey PixelSearch reference: https://autohotkey.com/docs/commands/PixelSearch.htm

        :param Union[str, int] color:
        :param int variation:
        :param Tuple[int, int] upper_bound:
        :param Optional[Tuple[int, int]] lower_bound:
        :param coord_mode
        :param fast:
        :param rgb:
        :return: the coordinates of the pixel; None if the pixel is not found
        """
        options = []
        if fast:
            options.append('Fast')
        if rgb:
            options.append('RGB')
        x1, y1 = upper_bound
        if lower_bound:
            x2, y2 = lower_bound
        else:
            x2, y2 = ('%A_ScreenWidth%', '%A_ScreenHeight%')

        script = self.render_template('screen/pixel_search.ahk',
                                      x1=x1, y1=y1, x2=x2, y2=y2,
                                      coord_mode=coord_mode,
                                      color=color,
                                      variation=variation,
                                      options=options)
        resp = self.run_script(script)
        try:
            return ast.literal_eval(resp)
        except SyntaxError:
            return None



from setuptools import setup
from io import open
test_requirements = ['behave', 'behave-classy', 'pytest']
extras = {'test': test_requirements}

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='ahk',
    version='0.5.0',
    url='https://github.com/spyoungtech/ahk',
    description='A Python wrapper for AHK',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author_email='spencer.young@spyoung.com',
    author='Spencer Young',
    packages=['ahk'],
    install_requires=['jinja2'],
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Desktop Environment',
        'Programming Language :: Python',
        'Environment :: Win32 (MS Windows)',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    tests_require=test_requirements,
    include_package_data=True,
    zip_safe=False
)
